#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

#if defined(WIN32) || defined(WIN64)
#include <windows.h>
void msleep(int msec) { Sleep( msec ); }
#else
#include <stdint.h>
#include <unistd.h>
#include <termios.h>
#include <errno.h>
void msleep(int msec) { usleep( msec*1000 ); }
int min(int a, int b) { return a < b ? a : b; }
#endif

#if defined(__APPLE__) && !defined(NOFRAMEWORK)
// include path to framework
#include <libxidcusb/xidcusb.h>
#else
#include <xidcusb.h>
#endif

void print_state (state_t* state)
{
	printf( "state: " );
	switch (state->device_state)
	{
		case DevStateOff: printf( "off" ); break;
		case DevStateStop: printf( "stop" ); break;
		case DevStateMoving: printf( "run" ); break;
		case DevStateTune: printf( "tune" ); break;
		case DevStateCalibr: printf( "calibrating" ); break;
		case DevStateAlarm: printf( "alarm" ); break;
		default: printf( "UNKNOWN" );
	}
	printf( "  pwm: " );
	switch (state->pwm_state)
	{
		case PwmStateOff      : printf( "off" ); break;
		case PwmStateBrak    	: printf( "brake" ); break;
		case PwmStateRunfwd	 	: printf( "runfwd" ); break;
		case PwmStateRunbck		: printf( "runbck" ); break;
		case PwmStateInvbrak	: printf( "invbrake" ); break;
		default: printf( "UNKNOWN" );
	}
	printf( " rpm: %d", state->current_rpm );
	printf( " pos: %d", state->current_position );
	printf( " temp: %d", state->current_temp );
	printf( " veng: %d", state->current_engine_voltage );
	printf( " vpwr: %d", state->current_power_voltage );
	printf( " currr: %d", state->current_input );
	printf( " flags: %x", state->flags );
	if (state->flags & STATE_ERRC)
		printf( " ERRC" );
	if (state->flags & STATE_ERRD)
		printf( " ERRD" );
	printf( "\n" );
}


void report_firmware (device_t device)
{
	result_t result;
	unsigned int firmware_ver[3];
	uint8_t is_firmware = 0;
	
	if ((result = has_firmware( device, &is_firmware )) != result_ok) { printf( "error command %d\n", result ); }
		
	if (is_firmware)
	{
		printf( "firmware is presented\n" );
		if ((result = get_firmware_version( device, firmware_ver+0, firmware_ver+1, firmware_ver+2 )) != result_ok)
			printf( "error command %d\n", result );
		printf( "firmware version: %d.%d.%d\n", firmware_ver[0], firmware_ver[1], firmware_ver[2] );
	}
	else
		printf( "firmware is not presented\n" );
}

uint32_t make_key (uint8_t b3, uint8_t b2, uint8_t b1, uint8_t b0)
{
	return ((uint32_t)b3 << 24) | ((uint32_t)b2 << 16) | ((uint32_t)b1 << 8) | b0;
}

uint32_t make_key_str (const char* s)
{
  uint32_t result;
  if (sscanf( s, "0x%x", &result ) == 1)
    return result;
  else if (sscanf( s, "%x", &result ) == 1)
    return result;
  else
    return 0;
}

int upload_firmware(char* device_name, char* firmware_file, char* key_str)
{
	device_t device;
	FILE* f;
	uint8_t* data;
	long len;
	result_t result;
	uint32_t skey;
	int upload_result = 0;

	printf( "opening device\n" );
	
	device = open_raw_device( device_name );
	if (device == device_undefined)
	{
		printf( "error opening raw device\n" );
		return 1;
	}

	msleep( 3*1000 );

	report_firmware( device );
	
	if (key_str)
	{	
		skey = make_key_str( key_str );
		printf( "Writing skey: %x\n", skey );

		if (write_key( device, skey ) != result_ok) { printf( "error command\n" ); }

		printf( "key is written\n" );

		//msleep( 3*1000 );
	}

	f = fopen( firmware_file, "rb" );
	if (!f)
	{
		printf( "Can't open file\n" );
		return 1;
	}
	fseek( f, 0, SEEK_END );
	len = ftell( f );
	rewind( f );
	data = malloc( len );
	if (len != (long)fread( data, 1, len, f ))
		return 1;
	printf( "Initiating upload\n" );
	if ((result = command_update_firmware( device, data, len )) != result_ok)
	{
		printf( "Error uploading firmware %d\n", result );
		upload_result = 1;
	}
	else
	{
		printf( "Firmware uploaded\n" );
	}
	free( data );
	fclose( f );

	if (close_device( device ) != result_ok) printf( "error closing\n" );
	return upload_result;
}

int main (int argc, char* argv[])
{
	char device_name[256];
	device_t device;
	state_t state;
	int names_count = 16;
	char** names;
	int i;
	int device_specified = 0;
	int key_specified = 0;
	char* firmware_file = NULL;
	char* key_str = NULL;

	printf( "Hello! I'm a firmware updater!\n" );
	printf( "I am %s bit\n", sizeof(int*)==4 ? "32" : "64" );
	printf( "Give %d arguments\n", argc-1 );
	for (i = 1; i < argc; ++i) printf( "  #%d: %s\n", i, argv[i] );

	if (argc > 4 && argc < 2)
	{
		printf( "usage: firmware firmware_file [device] [key]\n" );
		return 1;
	}
	firmware_file = argv[1];
	device_specified = argc >= 3;
	key_specified = argc >= 4;
	
	if (device_specified)
		strncpy( device_name, argv[2], sizeof(device_name) );
	if (key_specified)
		key_str = argv[3];
		
	if (enumerate_devices( &names_count, &names, 0 ) != result_ok)
	{
		printf( "error opening\n" );
		return 1;
	}

	if (names_count == 0)
	{
		printf( "No devices found" );
		if (!device_specified)
			return 1;
	}

	if (!device_specified)
		strncpy( device_name, names[0], sizeof(device_name) );

	for (i = 0; i < names_count; ++i)
	{
		printf( "device: %s\n", names[i] );
	}
	free_enumerate_devices( names_count, names );
	printf( "Using device: %s\n", device_name );

	printf( "Starting updating firmware\n" );
	if (upload_firmware( device_name, firmware_file, key_str ) == 0)
	{
		msleep( 10*1000 );

		// check out what we have done
		printf( "Check device\n" );

		device = open_device( device_name );
		if (device == device_undefined)
		{
			printf( "error opening\n" );
			return 1;
		}
	
		msleep( 3*1000 );

		report_firmware( device );
	
		if (get_status( device, &state ) != result_ok)
		{
			printf( "error command. maybe updating failed\n" );
		}
		else
		{
			print_state( &state );
			printf( "device seems to be ok!\n" );
		}
		
		if (close_device( device ) != result_ok) printf( "error closing\n" );
	}
	printf( "Done\n" );
	return 0;
}

